@@ -13,6 +13,7 @@ import signal
 import json
 import time
 
+from copy import deepcopy
 from os import path, makedirs
 from threading import Thread, Event
 
@@ -55,7 +56,7 @@ class Avatar(Thread):
         self.memory_ranges = intervaltree.IntervalTree()
         self.loaded_plugins = []
         self.cpu_model = cpu_model
-        
+
         if self.cpu_model is None and hasattr(self.arch, 'cpu_model'):
             self.cpu_model = self.arch.cpu_model
         # Setup output-dir and logging
@@ -88,7 +89,7 @@ class Avatar(Thread):
         self.queue = queue.Queue()
         self.fast_queue = queue.Queue()
         self.fast_queue_listener = AvatarFastQueueProcessor(self)
-        self.message_handlers = { 
+        self.message_handlers = {
             SyscallCatchedMessage: self._handle_syscall_catched_message,
             BreakpointHitMessage: self._handle_breakpoint_hit_message,
             UpdateStateMessage: self._handle_update_state_message,
@@ -107,7 +108,7 @@ class Avatar(Thread):
         :ivar file_name: (Absolute) path to the config file
         """
         if file_name is None:
-            file_name = "%s/conf.json" % self.output_directory 
+            file_name = "%s/conf.json" % self.output_directory
         with open(file_name, 'r') as config_file:
             config = json.load(config_file)
 
@@ -143,7 +144,7 @@ class Avatar(Thread):
         conf_dict['memory_mapping'] = []
         for mr in self.memory_ranges:
             conf_dict['memory_mapping'].append(mr.data.dictify())
-        
+
         conf_dict['targets'] = []
         for t in self.targets.values():
             conf_dict['targets'].append(t.dictify())
@@ -153,7 +154,7 @@ class Avatar(Thread):
 
     def save_config(self, file_name=None, config=None):
         if file_name is None:
-            file_name = "%s/conf.json" % self.output_directory 
+            file_name = "%s/conf.json" % self.output_directory
         conf_dict = self.generate_config() if config is None else config
         with open(file_name, "w") as conf_file:
             json.dump(conf_dict, conf_file)
@@ -177,9 +178,13 @@ class Avatar(Thread):
         self.log.info("Avatar Received SIGINT")
         self.sigint_handler()
 
-    def load_plugin(self, name):
-        plugin = __import__("avatar2.plugins.%s" % name,
-                            fromlist=['avatar2.plugins'])
+    def load_plugin(self, name, local=False):
+        if local is True:
+            plugin = __import__(name, fromlist=['.'])
+        else:
+            plugin = __import__("avatar2.plugins.%s" % name,
+                                fromlist=['avatar2.plugins'])
+
         plugin.load_plugin(self)
         self.loaded_plugins += [name]
 
@@ -224,11 +229,14 @@ class Avatar(Thread):
     def add_memory_range(self, address, size, name=None, permissions='rwx',
                          file=None, file_offset=None, file_bytes=None,
                          forwarded=False, forwarded_to=None, emulate=None,
-                         interval_tree=None, **kwargs):
+                         interval_tree=None, inline=False, overwrite=False,
+                         **kwargs):
         """
         Adds a memory range to avatar
 
         :param emulate:      Emulation function that will take name, address and size if set
+        :param inline:       If set to true, inline the emulation object into targets which support inlining (e.g., QEMU, PANDA).
+                             NB: This deactivates all callbacks/watchmen on this region.
         :param address:      Base-Address of the Range
         :param size:         Size of the range
         :param file:         A file backing this range, if applicable
@@ -238,9 +246,12 @@ class Avatar(Thread):
         :param forwarded_to: If forwarded is true, specify the forwarding target
         :param interval_tree:interval_tree this range shall be added to. If None,
                              the range will be added to self.memory_ranges
+        :param overwrite:    If true, overwrite existing MR if they overlap with the new one
         """
         memory_ranges = self.memory_ranges if interval_tree is None else interval_tree
-        if emulate:
+        if inline is True and emulate is None:
+            self.log.warn("inline set for non peripheral memory")
+        if emulate is not None:
             python_peripheral = emulate(name, address, size, **kwargs)
             forwarded = True
             forwarded_to = python_peripheral
@@ -248,11 +259,32 @@ class Avatar(Thread):
 
         if forwarded is True:
             kwargs.update({'qemu_name': 'avatar-rmemory'})
+        if inline is True:
+            kwargs.update({'qemu_name': 'avatar-pyperipheral'})
+            kwargs.update({'inline_module': str(python_peripheral.__module__)})
+
         m = MemoryRange(address, size, name=name, permissions=permissions,
                         file=file, file_offset=file_offset,
                         file_bytes=file_bytes, forwarded=forwarded,
                         forwarded_to=forwarded_to, **kwargs)
+
+
+        mr_set = memory_ranges[address:address+size]
+        if overwrite is True and len(mr_set) > 0:
+            start = min(mr_set, key=lambda x: x.begin).begin
+            end   = max(mr_set, key=lambda x: x.end).end
+            memory_ranges.chop(address, address+size,
+                                    datafunc=lambda x,y:
+                                    deepcopy(x.data) if y is True else x.data)
+            mr_set2 = memory_ranges[start:end]
+            for interval in mr_set2:
+                interval.data.address = interval.begin
+                interval.data.size = interval.end - interval.begin
+                interval.data.name = 'SPLIT_%x_%x' % (interval.data.address,
+                                                   interval.data.size)
+
         memory_ranges[address:address + size] = m
+
         return m
 
     def get_memory_range(self, address):
@@ -291,8 +323,8 @@ class Avatar(Thread):
         :type synced_ranges:   list
         """
 
-        if from_target.state != TargetStates.STOPPED or \
-                        to_target.state != TargetStates.STOPPED:
+        if from_target.state & TargetStates.STOPPED == 0 or \
+                        to_target.state & TargetStates.STOPPED == 0:
             raise Exception("Targets must be stopped for State Transfer, \
                              but target_states are (%s, %s)" %
                             (from_target.state, to_target.state))
@@ -303,7 +335,7 @@ class Avatar(Thread):
                 to_target.regs._get_names() & from_target.regs._get_names()
                 if hasattr(to_target, 'regs') and hasattr(from_target, 'regs')
                 else self.arch.registers)
-           
+
 
             # ARM may have banked registers; Apparantly, the order in which we
             # write them is important to QEMU and could to lead bugs otherwise.
@@ -359,13 +391,14 @@ class Avatar(Thread):
 
     @watch('RemoteMemoryRead')
     def _handle_remote_memory_read_message(self, message):
-        
+
         range = self.get_memory_range(message.address)
         if not range:
             return (message.id, None, False)
         message.dst = range
         if not range.forwarded:
-            raise Exception("Forward request for non forwarded range received!")
+            raise Exception("Forward request for non forwarded range received!\
+                            (Address = 0x%x)" % message.address)
         if range.forwarded_to is None:
             raise Exception("Forward request for non existing target received.\
                             (Address = 0x%x)" % message.address)
@@ -441,8 +474,8 @@ class Avatar(Thread):
 
 class AvatarFastQueueProcessor(Thread):
     """
-    The avatar fast queue handles events which require immediate action, 
-    i.e. TargetStateUpdates. 
+    The avatar fast queue handles events which require immediate action,
+    i.e. TargetStateUpdates.
     After processing, they get passed to the main avatar queue for further
     handling.
     """
