@@ -16,7 +16,7 @@ else:
 
 from avatar2.archs.arm import ARM
 from avatar2.targets import TargetStates
-from avatar2.message import AvatarMessage, UpdateStateMessage, BreakpointHitMessage
+from avatar2.message import AvatarMessage, UpdateStateMessage, BreakpointHitMessage, SyscallCatchedMessage
 
 GDB_PROT_DONE = 'done'
 GDB_PROT_CONN = 'connected'
@@ -92,11 +92,13 @@ class GDBResponseListener(Thread):
             pass  # library loading not supported yet
         elif msg == 'breakpoint-modified':
             pass  # ignore breakpoint modified for now
+        elif msg == 'breakpoint-created':
+            pass  # ignore breakpoint created for now
         elif msg == 'memory-changed':
             pass  # ignore changed memory for now
         elif msg == 'stopped':
             if payload.get('reason') == 'breakpoint-hit':
-                avatar_msg = BreakpointHitMessage(self._origin, payload['bkptno'],
+                avatar_msg = BreakpointHitMessage(self._origin, int(payload['bkptno']),
                                                   int(payload['frame']['addr'], 16))
             elif payload.get('reason') == 'exited-normally':
                 avatar_msg = UpdateStateMessage(
@@ -108,6 +110,9 @@ class GDBResponseListener(Thread):
                 if payload.get('signal-name') == 'SIGSEGV':
                     avatar_msg = UpdateStateMessage(
                         self._origin, TargetStates.EXITED)
+                elif payload.get('signal-name') == 'SIGTRAP':
+                    avatar_msg = BreakpointHitMessage(self._origin, -1,
+                                                      int(payload['frame']['addr'], 16))
                 else:
                     avatar_msg = UpdateStateMessage(
                         self._origin, TargetStates.STOPPED)
@@ -120,6 +125,12 @@ class GDBResponseListener(Thread):
             elif payload.get('reason') == 'read-watchpoint-trigger':
                 avatar_msg = UpdateStateMessage(
                     self._origin, TargetStates.STOPPED)
+            elif payload.get('reason') == 'syscall-entry':
+                avatar_msg = SyscallCatchedMessage(self._origin, int(payload['bkptno']),
+                                                  int(payload['frame']['addr'], 16), 'entry')
+            elif payload.get('reason') == 'syscall-return':
+                avatar_msg = SyscallCatchedMessage(self._origin, int(payload['bkptno']),
+                                                  int(payload['frame']['addr'], 16), 'return')
             elif payload.get('reason') is not None:
                 self.log.critical("Target stopped with unknown reason: %s" %
                                   payload['reason'])
@@ -180,7 +191,7 @@ class GDBResponseListener(Thread):
 
             try:
                 responses = self._gdb_controller.get_gdb_response(
-                    timeout_sec=0.0
+                    timeout_sec=0.01
                 )
             except:
                 continue
@@ -249,19 +260,25 @@ class GDBProtocol(object):
         gdb_args = []
         if not enable_init_files:
             gdb_args += ['--nx']
-        gdb_args = ['--quiet', '--interpreter=mi2']
+        gdb_args += ['--quiet', '--interpreter=mi2']
         gdb_args += additional_args
         if binary is not None:
             gdb_args += ['--args', binary]
             if local_arguments is not None:
                 gdb_args += [local_arguments]
-        self._gdbmi = pygdbmi.gdbcontroller.GdbController(
-            gdb_path=gdb_executable,
-            gdb_args=gdb_args,
-            verbose=verbose)  # set to True for debugging
+
+
+        if sys.version_info <= (3, 5):
+            self._gdbmi = pygdbmi.gdbcontroller.GdbController(
+                gdb_path=gdb_executable,
+                gdb_args=gdb_args,
+                verbose=verbose)  # set to True for debugging
+        else:
+            self._gdbmi = pygdbmi.gdbcontroller.GdbController(
+                command=[gdb_executable] + gdb_args,
+                time_to_check_for_additional_output_sec=0)
         queue = avatar.queue if avatar is not None else None
         fast_queue = avatar.fast_queue if avatar is not None else None
-
         self._communicator = GDBResponseListener(
             self, self._gdbmi, queue, fast_queue, origin)
         self._communicator.daemon = True
@@ -283,7 +300,7 @@ class GDBProtocol(object):
             self._gdbmi.exit()
             self._gdbmi = None
 
-    def _sync_request(self, request, rexpect):
+    def _sync_request(self, request, rexpect, timeout=5):
         """ Generic method to send a syncronized request
 
         :param request: the request as list
@@ -300,7 +317,7 @@ class GDBProtocol(object):
 
         self._gdbmi.write(req, read_response=False, timeout_sec=0)
         try:
-            response = self._communicator.get_sync_response(token)
+            response = self._communicator.get_sync_response(token, timeout=timeout)
             ret = True if response['message'] == rexpect else False
         except:
             response = None
@@ -352,6 +369,15 @@ class GDBProtocol(object):
         if self._arch.gdb_name == 'arm':
             self.set_abi('AAPCS')
 
+        if hasattr(self._arch, 'endian'):
+            req = ['-gdb-set', 'endian', self._arch.endian]
+            ret, resp = self._sync_request(req, GDB_PROT_DONE)
+            if not ret:
+                self.log.critical(
+                    "Unable to set endianness, received response: %s" %
+                    resp)
+                raise Exception("GDBProtocol was unable to set endianness")
+
         req = ['-target-select', 'remote', '%s:%d' % (ip, int(port))]
         ret, resp = self._sync_request(req, GDB_PROT_CONN)
 
@@ -361,7 +387,7 @@ class GDBProtocol(object):
         if not ret:
             self.log.critical("GDBProtocol was unable to connect to remote target")
             raise Exception("GDBProtocol was unable to connect")
-        
+
         self.update_target_regs()
 
         return ret
@@ -470,6 +496,8 @@ class GDBProtocol(object):
         :param bool temporary:  Tempory breakpoint
         :param str regex:     If set, inserts breakpoints matching the regex
         :param str condition: If set, inserts a breakpoint with specified condition
+                              Note: This allows only a single condition. For more
+                                    complex ones, please use set_break_condition()
         :param int ignore_count: Amount of times the bp should be ignored
         :param int thread:    Threadno in which this breakpoints should be added
         :returns:             The number of the breakpoint
@@ -504,6 +532,18 @@ class GDBProtocol(object):
         self.log.debug("Attempted to set breakpoint. Received response: %s" % resp)
         return int(resp['payload']['bkpt']['number']) if ret else -1
 
+    def set_break_condition(self, bp_no, condition='1'):
+        """
+        Modifies the break condition for a given break-point
+        :param bp_no: the breakpoint whose condition should be modified
+        :param condition: The condition string to use.
+                          The default value, '1', will make the bp unconditional
+        """
+        cmd = ["-break-condition", str(bp_no), condition]
+        ret, resp = self._sync_request(cmd, GDB_PROT_DONE)
+        self.log.debug("Attempted to set break-condition. Received response: %s" % resp)
+        return ret
+
     def set_watchpoint(self, variable, write=True, read=False):
         cmd = ["-break-watch"]
         if read is True and write is True:
@@ -536,6 +576,26 @@ class GDBProtocol(object):
         else:
             return -1
 
+    def set_syscall_cachpoint(self, syscall):
+        '''
+        Set's up a syscall catchpoint.
+        :param syscall: the syscall to catch; can be either its name or no
+        '''
+        # this command is not exported via gdb-mi, hence we need manual parsing
+        ret = self.console_command('catch syscall %s' % syscall)
+        if ret[0] is not True:
+            return ret[0]
+        # assume return message is in format of
+        # '"\nCatchpoint 1 (syscall 'write' [4])\\n") '
+        expected_bp_num = ret[1].split()[1]
+        if not expected_bp_num.isdigit():
+            self.log.warning("Couldn't extract bp_num for catchpoint!")
+            return True
+        return int(expected_bp_num)
+
+
+
+
     def remove_breakpoint(self, bkpt):
         """Deletes a breakpoint"""
         ret, resp = self._sync_request(
@@ -591,7 +651,7 @@ class GDBProtocol(object):
         """reads memory
 
         :param address:   Address to write to
-        :param wordsize:  the size of a read word (1, 2, 4 or 8) 
+        :param wordsize:  the size of a read word (1, 2, 4 or 8)
         :param num_words: the amount of read words
         :param raw:       Whether the read memory should be returned unprocessed
         :return:          The read memory
@@ -634,7 +694,7 @@ class GDBProtocol(object):
             return self._read_special_reg_from_name(reg)
         else:
             reg_nr = (
-                self._origin.regs._get_nr_from_name(reg) 
+                self._origin.regs._get_nr_from_name(reg)
                 if hasattr(self._origin, 'regs')
                 else self._arch.registers[reg])
             return self.read_register_from_nr(reg_nr)
@@ -688,7 +748,7 @@ class GDBProtocol(object):
 
             fmt = "{:s}=" \
                   + self._arch.special_registers[reg]['format'].replace(' ','')
-            
+
             ret, resp = self._sync_request(
                 ["-data-evaluate-expression", fmt.format(
                    self._arch.special_registers[reg]['gdb_expression'], *value)
@@ -744,6 +804,26 @@ class GDBProtocol(object):
             resp)
         return ret
 
+    def set_file(self, elf=''):
+        """Load an ELF file
+        :returns: True on success"""
+        ret, resp = self._sync_request(["-file-exec-and-symbols", elf], GDB_PROT_DONE)
+
+        self.log.debug(
+            "Attempted to load elf file. Received response: %s" %
+            resp)
+        return ret
+
+    def download(self):
+        """Download code to target
+        :returns: True on success"""
+        ret, resp = self._sync_request(["-target-download"], GDB_PROT_DONE, timeout=60)
+
+        self.log.debug(
+            "Attempted to download code to target. Received response: %s" %
+            resp)
+        return ret
+
     def set_endianness(self, endianness='little'):
         req = ['-gdb-set', 'endian', '%s' % endianness]
         ret, resp = self._sync_request(req, GDB_PROT_DONE)
@@ -779,7 +859,7 @@ class GDBProtocol(object):
             regex = re.compile("(0x[0-9a-f]*)[ .]")
             resp = regex.findall(resp)
             if len(resp) == 1:
-                resp = long(resp[0], 16)
+                resp = int(resp[0], 16)
             else:
                 resp = -1
                 ret = False
@@ -797,3 +877,8 @@ class GDBProtocol(object):
             self.log.debug("Unable to set variable %s to %s" %
                            (str(variable), str(value)))
         return ret
+
+    def quit(self):
+        req = ['-gdb-exit']
+        ret = self._sync_request(req, GDB_PROT_DONE)
+        return ret
