@@ -1,6 +1,9 @@
 import logging
+import time
+from sys import version_info
 from functools import wraps
 from threading import Event
+from types import MethodType
 
 from enum import IntEnum
 
@@ -13,7 +16,7 @@ def action_valid_decorator_factory(state, protocol):
     requested actions on a target, such as step(), stop(), read_register(), 
     write_register() and so on are actually executable.
 
-    :param state: The required state of the Target
+    :param state: A mask specifying the required state of the Target
     :type state:  An entry of the Enum TargetStates
     :param protocol: The protocol required to execute the action.
     :type protocol: str
@@ -63,6 +66,7 @@ def synchronize_state(*states, **kwargs):
                     if message.origin == self:
                         if message.state == state:
                             state_reached.set()
+                            avatar.watchmen.remove_watchman('UpdateState', w)
                         elif message.state == TargetStates.EXITED:
                             raise Exception("Target %s exited" % self.name)
 
@@ -75,7 +79,6 @@ def synchronize_state(*states, **kwargs):
             if blocking is True:
                 if not (transition_optional == True and self.state == state):
                     state_reached.wait()
-                avatar.watchmen.remove_watchman('UpdateState', w)
             return ret
 
         return state_synchronizer
@@ -94,6 +97,7 @@ class TargetStates(IntEnum):
     SYNCING = 0x10
     EXITED = 0x20
     NOT_RUNNING = INITIALIZED | STOPPED
+    BREAKPOINT = SYNCING | STOPPED
 
 class TargetRegs(object):
     def __init__(self, target, register_dict):
@@ -153,6 +157,7 @@ class TargetProtocolStore(object):
     def shutdown(self):
         """Shutsdown all the associated protocols"""
         for p in self.protocols:
+            #print("Unloading %s" % str(p))
             setattr(self, p, None)
 
     def __setattr__(self, name, value):
@@ -209,7 +214,6 @@ class Target(object):
         self.protocols = TargetProtocolStore()
 
         self.state = TargetStates.CREATED
-        self._no_state_update_pending = Event()
 
         self.log = logging.getLogger('%s.targets.%s' % (avatar.log.name, self.name))
         log_file = logging.FileHandler('%s/%s.log' % (avatar.output_directory, self.name))
@@ -219,6 +223,33 @@ class Target(object):
 
         self.regs = TargetRegs(self, self._arch.registers)
 
+    def dictify(self):
+        """
+        Returns the memory range as *printable* dictionary for the config
+        """
+
+        ignore = ['state', 'status', 'regs', 'protocols', 'log', 'avatar']
+        ignored_types = (MethodType)
+        expected_types = (str, bool, int, list) 
+        if version_info < (3, 0): expected_types += (unicode, )
+
+        t_dict = {'type': self.__class__.__name__,
+                  'module': self.__module__}
+
+        for k, v in self.__dict__.items():
+            if k in ignore: continue
+            if k.startswith('_'): continue
+            if v is None: continue
+            if isinstance(v, ignored_types): continue
+            if not isinstance(v, expected_types):
+                raise Exception(
+                    "Unsupported type %s for dictifying %s for target %s" %
+                    (type(v), k, self.name))
+            t_dict[k] = v
+        return t_dict
+
+
+
     @watch('TargetInit')
     def init(self):
         """
@@ -233,8 +264,6 @@ class Target(object):
         """
         self.protocols.shutdown()
 
-
-
     @watch('TargetCont')
     @action_valid_decorator_factory(TargetStates.STOPPED, 'execution')
     @synchronize_state(TargetStates.RUNNING)
@@ -250,10 +279,6 @@ class Target(object):
     @action_valid_decorator_factory(TargetStates.RUNNING, 'execution')
     @synchronize_state(TargetStates.STOPPED, transition_optional=True)
     def stop(self, blocking=True):
-        """
-        Stops the execution of the target 
-        :param blocking: if True, block until the target is STOPPED
-        """
         return self.protocols.execution.stop()
 
     @watch('TargetStep')
@@ -266,8 +291,48 @@ class Target(object):
         """
         return self.protocols.execution.step()
 
-    @watch('TargetWriteMemory')
+    @watch('TargetSetFile')
+    @action_valid_decorator_factory(TargetStates.STOPPED, 'execution')
+    def set_file(self, elf):
+        """
+        Load an ELF file
+
+        :param elf: ELF file to load
+        :returns: True on success else False
+        """
+        if not hasattr(self.protocols.execution, 'set_file'):
+            self.log.error('Protocol "' + type(self.protocols.execution).__name__ + '" does not support "set_file"')
+            return False
+
+        return self.protocols.execution.set_file(elf)
+
+    @watch('TargetDownload')
+    @action_valid_decorator_factory(TargetStates.STOPPED, 'execution')
+    def download(self):
+        """
+        Download the loaded code to the Target
+
+        :returns: True on success else False
+        """
+        if not hasattr(self.protocols.execution, 'download'):
+            self.log.error('Protocol "' + type(self.protocols.execution).__name__ + '" does not support "download"')
+            return False
+
+        return self.protocols.execution.download()
+
+    @watch('TargetGetSymbol')
     @action_valid_decorator_factory(TargetStates.STOPPED, 'memory')
+    def get_symbol(self, symbol):
+        """
+        Get the address of a symbol
+
+        :param symbol:    The name of a symbol whose address is wanted
+        :returns:         (True, Address) on success else False
+        """
+        return self.protocols.memory.get_symbol(symbol)
+
+    @watch('TargetWriteMemory')
+    #@action_valid_decorator_factory(TargetStates.STOPPED, 'memory')
     def write_memory(self, address, size, value, num_words=1, raw=False):
         """
         Writing to memory of the target
@@ -283,25 +348,45 @@ class Target(object):
         :param raw:       Specifies whether to write in raw or word mode
         :returns:         True on success else False
         """
+        try:
+            target_range = self.avatar.get_memory_range(address)
+        except Exception as e:
+            self.log.warn("Performing write on undefined range at 0x%x" % address)
+            target_range = None
+        if target_range is not None and target_range.forwarded is True and \
+        target_range.forwarded_to != self:
+            return target_range.forwarded_to.write_memory(address, size, value,
+                                                          num_words, raw)
+        
         return self.protocols.memory.write_memory(address, size, value,
                                                   num_words, raw)
 
     @watch('TargetReadMemory')
     @action_valid_decorator_factory(TargetStates.STOPPED, 'memory')
-    def read_memory(self, address, size, words=1, raw=False):
+    def read_memory(self, address, size, num_words=1, raw=False):
         """
         Reading from memory of the target
 
         :param address:     The address to read from 
         :param size:        The size of a read word
-        :param words:       The amount of words to read (default: 1)
+        :param num_words:   The amount of words to read (default: 1)
         :param raw:         Whether the read memory is returned unprocessed
         :return:          The read memory
         """
-        return self.protocols.memory.read_memory(address, size, words, raw)
+        try:
+            target_range = self.avatar.get_memory_range(address)
+        except Exception as e:
+            self.log.warn("Performing write on undefined range at 0x%x" % address)
+            target_range = None
+        if target_range is not None and target_range.forwarded is True and \
+        target_range.forwarded_to != self:
+            return target_range.forwarded_to.read_memory(address, size,
+                                                         num_words, raw)
+        
+        return self.protocols.memory.read_memory(address, size, num_words, raw)
 
     @watch('TargetRegisterWrite')
-    @action_valid_decorator_factory(TargetStates.STOPPED, 'registers')
+    #@action_valid_decorator_factory(TargetStates.STOPPED, 'registers')
     def write_register(self, register, value):
         """
         Writing a register to the target
@@ -312,7 +397,7 @@ class Target(object):
         return self.protocols.registers.write_register(register, value)
 
     @watch('TargetRegisterRead')
-    @action_valid_decorator_factory(TargetStates.STOPPED, 'registers')
+    #@action_valid_decorator_factory(TargetStates.STOPPED, 'registers')
     def read_register(self, register):
         """
         Reading a register from the target
@@ -364,12 +449,19 @@ class Target(object):
     def update_state(self, state):
         self.log.info("State changed to to %s" % TargetStates(state))
         self.state = state
-        #self._no_state_update_pending.set()
 
     @watch('TargetWait')
-    def wait(self, state=TargetStates.STOPPED):
-        while self.state & state == 0:
-            pass
+    def wait(self, state=TargetStates.STOPPED|TargetStates.EXITED):
+
+        if state & TargetStates.SYNCING:
+            self.log.warn("Waiting on SYNCING-state - this could lead to races")
+
+        while True:
+            if self.state & state != 0 and \
+               (state & TargetStates.SYNCING or \
+                self.state & TargetStates.SYNCING == 0):
+                break
+            time.sleep(.001) # send thread a ms to sleep to free resources
 
     def get_status(self):
         """
